/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.valueprovider;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.EditingSupport;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.AnnotationEntry;
import org.fortiss.tooling.base.ui.annotation.editingsupport.CheckBoxEditingSupport;
import org.fortiss.tooling.base.ui.annotation.editingsupport.ComboBoxEditingSupport;
import org.fortiss.tooling.base.ui.annotation.editingsupport.TextEditingSupport;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * Base class for {@link IAnnotationValueProvider}s.
 * 
 * @author diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public abstract class ValueProviderBase<T extends IAnnotatedSpecification> implements
		IAnnotationValueProvider<T> {

	/**
	 * {@link EClass} of {@link IAnnotatedSpecification} that contains the respective structural
	 * feature and that is managed by this {@link IAnnotationValueProvider}.
	 */
	protected final EClass annotatedSpecificationEClass;

	/**
	 * Map: instanceKey -> Fixed input choice to be offered in a combo box. The semantics of the map
	 * value is as follows:
	 * <ul>
	 * <li> {@code fixedInputChoice[key] == null} -> do not use fixed values, provide text input
	 * cell.</li>
	 * <li> {@code !fixedInputChoice[key].isEmpty()} -> provide a fixed set of choices in an input
	 * box.
	 * <li> {@code fixedInputChoice[key].isEmpty()} -> dynamically manage the set of provided values,
	 * i.e. the combo box is editable and enables the user to add new values and remove existing
	 * choices (by removing all instances of them).</li>
	 * <li>For enum types, the set of inputs is automatically initialized from the underlying
	 * {@link EEnum} type definition.</li>
	 */
	private Map<String, List<String>> fixedInputChoiceMap;

	/**
	 * Map: instanceKey -> Current input choice, computed according to the initialization of
	 * {@link #fixedInputChoiceMap}.
	 */
	private Map<String, Collection<String>> currentInputChoiceMap;

	/**
	 * Constructs a new {@link IAnnotationValueProvider} for an annotation specified in a given
	 * {@link EClass}.
	 */
	public ValueProviderBase(EClass annotatedSpecificationEClass) {
		this.annotatedSpecificationEClass = annotatedSpecificationEClass;

		fixedInputChoiceMap = new HashMap<String, List<String>>();
		currentInputChoiceMap = new HashMap<String, Collection<String>>();
	}

	/** {@inheritDoc} */
	@Override
	public boolean setAnnotationName(String name, T specification) {
		// Renaming this annotation is not supported.
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean canEdit(T specification) {
		return canEdit(specification, DEFAULT_KEY);
	}

	/** {@inheritDoc} */
	@Override
	public <U> U getAnnotationValue(T specification, String instanceKey) throws Exception {
		return getAnnotationValue(specification);
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification) throws Exception {
		setAnnotationValue(value, specification, null);
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification) throws Exception {
		setAnnotationValue(value, specification, null);
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification, String instanceKey)
			throws Exception {
		throw new Exception(
				"setAnnotationValue(String) is not supported / has not been implemented for annotations of type " +
						specification.getClass());
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification, String instanceKey)
			throws Exception {
		throw new Exception(
				"setAnnotationValue(U) is not supported / has not been implemented for annotations of type " +
						specification.getClass());
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws Exception
	 */
	@Override
	public EditingSupport createEditingSupport(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, T specification, String instanceKey)
			throws Exception {

		try {
			initializeInputChoice(viewer, clazz, instanceKey);
		} catch(Exception e) {
			// Use standard text editor as fallback
			currentInputChoiceMap = null;
		}

		// Input is not restricted -> free editing in text cell
		if(currentInputChoiceMap.get(instanceKey) == null) {

			if(getAnnotationValue(specification, instanceKey) instanceof Boolean) {
				return new CheckBoxEditingSupport(viewer, clazz, instanceKey);
			}

			return new TextEditingSupport(viewer, clazz, instanceKey);
		}

		// Multiple instances are not supported (yet) for input to be edited in Comboboxes
		assert (instanceKey == null);

		// Input is restricted to concrete set of values
		// (use SingleEnumAttributeValueProviderBase for Enum types!)
		return new ComboBoxEditingSupport(viewer, clazz, instanceKey,
				currentInputChoiceMap.get(instanceKey), fixedInputChoiceMap.get(instanceKey)
						.isEmpty());
	}

	/**
	 * Creates the editing support without referring to the optional {@code instanceKey} (see
	 * {@link #createEditingSupport(ColumnViewer, Class, IAnnotatedSpecification, String)}.
	 */
	public EditingSupport createEditingSupport(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz) throws Exception {

		return createEditingSupport(viewer, clazz, null, null);
	}

	/** {@inheritDoc} */
	@Override
	public List<Class<? extends EObject>> excludeModelElementTypeFromAnnotatedSpecification() {
		return null;
	}

	/**
	 * Add decorations to an annotation when its instance is newly created by
	 * {@link #getAnnotatedSpecificationForModelElement(IModelElement)}.
	 * 
	 * @param specification
	 *            annotation to be decorated
	 */
	protected void decorateAnnotationSpecification(T specification) throws Exception {
		decorateAnnotationSpecification(specification, null);
	}

	/**
	 * Add decorations to an annotation when its instance is newly created by
	 * {@link #getAnnotatedSpecificationForModelElement(IModelElement)}.
	 * 
	 * @param specification
	 *            annotation to be decorated
	 * @param instanceKey
	 *            instanceKey of annotation to be decorated
	 */
	protected void decorateAnnotationSpecification(T specification, String instanceKey)
			throws Exception {
		// No decoration
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public T getAnnotatedSpecificationForModelElement(IModelElement element) throws Exception {

		T rval =
				(T)pickFirstInstanceOf(
						(Class<IAnnotatedSpecification>)annotatedSpecificationEClass
								.getInstanceClass(),
						element.getSpecifications());

		if(rval != null) {
			return rval;
		}

		// Create annotation
		T specification =
				(T)annotatedSpecificationEClass.getEPackage().getEFactoryInstance()
						.create(annotatedSpecificationEClass);
		if(getInstanceKeys(specification).isEmpty()) {
			decorateAnnotationSpecification(specification);
		} else {
			for(String key : getInstanceKeys(specification)) {
				try {
					decorateAnnotationSpecification(specification, key);
				} catch(Exception e) {
					// Some of the keys might not be backed by an implementation that needs to
					// "decorate" the underlying specification object (e.g., a custom key that
					// represents a calculated, volatile value. Hence, just ignore invalid keys
					// during this setup phase. The other methods (get/set/...) will throw an
					// Exception on invalid keys.
				}
			}
		}

		// Hook specification to model element
		element.addSpecification(specification);

		return specification;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Class<T> getAnnotationClazz() {
		return (Class<T>)annotatedSpecificationEClass.getInstanceClass();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<String> getInstanceKeys(T specification) {
		return Collections.emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean allowsDynamicAnnotationInstances() {
		return false;
	}

	/** Sets the {@link #fixedInputChoiceMap} field for a specific instance key */
	public void setFixedInputChoice(List<String> fixedInputChoice, String instanceKey) {
		this.fixedInputChoiceMap.put(instanceKey, fixedInputChoice);
	}

	/**
	 * Sets the {@link #fixedInputChoiceMap} field for an {@link IAnnotationValueProvider} that
	 * manages only a single instance key.
	 */
	public void setFixedInputChoice(List<String> fixedInputChoice) {
		setFixedInputChoice(fixedInputChoice, DEFAULT_KEY);
	}

	/**
	 * Initializes the {@link #currentInputChoiceMap} field (called during the construction of the
	 * editing support in {@link #createEditingSupport(ColumnViewer, Class)}).
	 */
	@SuppressWarnings("unchecked")
	private void initializeInputChoice(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, String instanceKey) throws Exception {
		// Input choice is not set, or it is restricted to concrete set of values
		// (use SingleEnumAttributeValueProviderBase for Enum types!)
		if(fixedInputChoiceMap.get(instanceKey) == null ||
				!fixedInputChoiceMap.get(instanceKey).isEmpty()) {
			currentInputChoiceMap.put(instanceKey, fixedInputChoiceMap.get(instanceKey));

		} else if(!(viewer.getInput() instanceof Set<?>) || ((Set<?>)viewer.getInput()).isEmpty() ||
				!(((Set<?>)viewer.getInput()).iterator().next() instanceof AnnotationEntry)) {
			currentInputChoiceMap.put(instanceKey, null);

		} else {

			AnnotationEntry annotationEntry =
					((Set<AnnotationEntry>)viewer.getInput()).iterator().next();

			EObject root = annotationEntry.getModelElement();
			while(root.eContainer() != null) {
				root = root.eContainer();
			}

			Set<String> choices = new TreeSet<String>();
			for(T specification : (EList<T>)EcoreUtils.getChildrenWithType(root, clazz)) {
				choices.add(getAnnotationValue(specification).toString());
			}

			currentInputChoiceMap.put(instanceKey, choices);
		}
	}

	/**
	 * Updates the current input choice, and returns the actual value. Called from
	 * {@link #setAnnotationValue(String, IAnnotatedSpecification)}. In particular, this
	 * method handles the removal of input choices from if these are dynamically managed.
	 */
	protected String getAnnotationValueAndUpdateInputChoice(String value, T specification,
			String instanceKey) throws Exception {
		if((value == null || value.isEmpty()) && currentInputChoiceMap.get(instanceKey) != null &&
				!currentInputChoiceMap.get(instanceKey).isEmpty()) {
			// Only one entry left: do not remove
			if(currentInputChoiceMap.get(instanceKey).size() == 1) {
				return currentInputChoiceMap.get(instanceKey).iterator().next();
			}

			// 1st entry is to be removed: Choose 2nd entry as new value
			String currentValue = getAnnotationValue(specification);
			if(currentInputChoiceMap.get(instanceKey).iterator().next().equals(currentValue)) {
				Iterator<String> iter = currentInputChoiceMap.get(instanceKey).iterator();
				iter.next();
				String rval = iter.next();
				currentInputChoiceMap.get(instanceKey).remove(currentValue);
				return rval;
			}

			// Otherwise, return the entry just before the entry to be removed.
			String rval = null;
			for(Iterator<String> iter = currentInputChoiceMap.get(instanceKey).iterator(); iter
					.hasNext();) {
				String s = iter.next();
				if(s != null && s.equals(currentValue)) {
					currentInputChoiceMap.get(instanceKey).remove(currentValue);
					return rval;
				}
				rval = s;
			}
		}
		return null;
	}
}
