/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view.generic;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.ColumnViewerToolTipSupport;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.IElementComparer;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.window.ToolTip;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.AnnotationEntry;
import org.fortiss.tooling.base.ui.annotation.IAnnotationValueService;
import org.fortiss.tooling.base.ui.annotation.editingsupport.ElementCommentEditingSupport;
import org.fortiss.tooling.base.ui.annotation.editingsupport.ElementNameEditingSupport;
import org.fortiss.tooling.base.ui.annotation.labelprovider.AnnotationLabelProvider;
import org.fortiss.tooling.base.ui.annotation.labelprovider.ElementCommentLabelProvider;
import org.fortiss.tooling.base.ui.annotation.labelprovider.ElementNameLabelProvider;
import org.fortiss.tooling.base.ui.annotation.labelprovider.EmulatedNativeCheckBoxLabelProvider;
import org.fortiss.tooling.base.ui.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.base.ui.annotation.view.AnnotationViewPartBase;
import org.fortiss.tooling.base.ui.annotation.view.generic.filter.AnnotationFilter;
import org.fortiss.tooling.base.ui.annotation.view.generic.filter.AnnotationFilterWidget;

/**
 * <p>
 * This class provides a generic annotation view whose methods are intended to be reused by child
 * classes. Inherited classes can avoid code duplication, e.g. the creation of the table and its
 * first row, using this class.
 * </p>
 * <p>
 * In order to support displaying and editing specific annotation types, the corresponding
 * {@link IAnnotationValueProvider#createEditingSupport(org.eclipse.jface.viewers.ColumnViewer, Class, IAnnotatedSpecification, String)}
 * factory method must return an appropriate {@link EditingSupport}.
 * </p>
 * 
 * @author eder, diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: D91763408246ECD3578DAD0E37BC4C41
 */
public class GenericAnnotationView extends AnnotationViewPartBase {
	/** Root composite of {@link GenericAnnotationView}. */
	Composite rootComposite;

	/** The table viewer */
	protected TableViewer tableViewer;

	/** Row an column filter for annotations */
	private AnnotationFilter annotationFilter;

	/** Widget for filter options */
	private AnnotationFilterWidget annotationFilterWidget;

	/**
	 * Last set of {@link AnnotationEntry}s used to {@link #update(Collection)} the
	 * {@link GenericAnnotationView}, i.e. during the construction of the, or because of a model
	 * change. The reference is required to update view after a change of the column (annotation)
	 * filter.
	 */
	private Collection<AnnotationEntry> lastAnnotationEntries;

	/** Number of fixed columns shown for every model element (e.g., name, comment). */
	int fixedColumnCount;

	/** Map used to preserve column width during update of the view. */
	Map<String, Integer> columnWidthCache;

	/** Default width of columns */
	static final int COLUMN_DEFAULT_WIDTH = 125;

	/**
	 * Sets a fixed {@link AnnotationFilter}.
	 * <ul>
	 * <li>If the {@link AnnotationFilter} is non-{@code null}, it sets the filter and disables the
	 * {@link AnnotationFilterWidget}. For instance, this can be used to embed / extend the
	 * {@link GenericAnnotationView} into a specialized view dedicated to a particular model type.</li>
	 * <li>Otherwise, this method reinitializes the {@link AnnotationFilterWidget}.</li>
	 * </ul>
	 */
	public void setFixedAnnotationFilter(AnnotationFilter annotationFilter) {

		// Remove old table row filter
		tableViewer.removeFilter(this.annotationFilter);

		if(annotationFilter != null) {
			// Externally created filter -> remove filter option GUI
			this.annotationFilter = annotationFilter;
			if(annotationFilterWidget != null) {
				annotationFilterWidget.dispose();
				annotationFilterWidget = null;
			}
		} else {
			// No externally created filter -> setup filter option GUI
			this.annotationFilter = new AnnotationFilter();
			setupAnnotationFilterWidget();
		}

		// Add new table row filter
		tableViewer.addFilter(this.annotationFilter);

		// Update layout
		rootComposite.layout(true, true);
	}

	/** Returns the view's {@link AnnotationFilter}. */
	public AnnotationFilter getAnnotationFilter() {
		return annotationFilter;
	}

	/** Computes the level of an {@link IModelElement} in the tree */
	public static int getModelElementLevel(EObject o) {
		int level = 0;
		while(o.eContainer() != null) {
			level++;
			o = o.eContainer();
		}

		return level;
	}

	/**
	 * Returns the ancestor of the {@link IModelElement} {@code modelElement} that is
	 * {@code levelsUp} levels above model element in the hierarchy, or {@code null}, in case
	 * {@code levelsUp} > depth(modelElement}.
	 */
	public static EObject getModelElementAncestor(EObject o, int levelsUp) {
		while(o.eContainer() != null && levelsUp > 0) {
			o = o.eContainer();
			levelsUp--;
		}
		return o;

	}

	/** Updates the table after a filter change */
	public void update() {
		// Full update, i.e. including reconstructions of columns is always required due to the
		// following scenario:
		// - Filter is set for annotation names.
		// - Filter scope is changed to component names -> annotation filter needs to be undone
		// - Filter is based on annotation entry type
		// - Filter is set to current level / current sub model
		if(lastAnnotationEntries != null) {
			update(lastAnnotationEntries);
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void update(Collection<AnnotationEntry> annotationEntries) {
		// Update row filter with currently selected element
		annotationFilter.setCurrentySelectedModelElement(getCurrentlySelectedObject());

		if(annotationFilter.annotationViewQueryUpdateRequired() || lastAnnotationEntries == null ||
				!annotationEntries.equals(lastAnnotationEntries)) {
			// Update filter widget
			if(annotationFilterWidget != null) {
				annotationFilterWidget.updateAnnotationTypeFilterOptionsComboBox(annotationEntries,
						this);
			}

			// Remember last annotation entries (required to re-draw table after changing the
			// column/annotation filter
			lastAnnotationEntries = annotationEntries;

			// Minimize flickering while updating the table
			tableViewer.getTable().setRedraw(false);

			// Input must be set before the construction of the columns, since the editing support
			// for a given cell might need to query the contents of other model elements (e.g., in
			// order to aggregate a set of admissible choices for its value).
			tableViewer.setInput(annotationEntries);

			// Sorted set of {@link ColumnHandle} used to instantiate this {@link
			// GenericAnnotationView}'s columns in the right order.
			Set<ColumnHandle> sortedColumnHandles = new TreeSet<ColumnHandle>();

			// Dispose all columns of the table viewer, except the leading fixed columns, which e.g.
			// hold the model elements' names and comments.
			while(tableViewer.getTable().getColumnCount() > fixedColumnCount) {
				TableColumn column = tableViewer.getTable().getColumn(fixedColumnCount);
				columnWidthCache.put(column.getText(), column.getWidth());
				column.dispose();
			}

			// Aggregate required columns. Column order is defined by ColumnHandle.compareTo().
			for(AnnotationEntry entry : annotationEntries) {
				for(IAnnotatedSpecification spec : entry.getSpecificationsList()) {
					// Create new column for each instance of the current multi-instance annotation
					if((entry.allowsDynamicAnnotationInstances(spec.getClass())) ||
							!entry.getInstanceKeys(spec.getClass()).isEmpty()) {

						for(String instanceKey : entry.getInstanceKeys(spec.getClass())) {
							ColumnHandle columnHandle =
									new ColumnHandle(entry, spec, instanceKey, false);
							if(annotationFilter.passesColumnFilter(columnHandle)) {
								sortedColumnHandles.add(columnHandle);
							}
						}
					}

					// Create a new column for
					// - ordinary (single-instance) annotations
					// - "create" column for multi-instance annotations for which instances can be
					// created dynamically
					if(entry.getInstanceKeys(spec.getClass()).isEmpty() ||
							entry.allowsDynamicAnnotationInstances(spec.getClass())) {
						ColumnHandle columnHandle =
								new ColumnHandle(entry, spec, null,
										entry.allowsDynamicAnnotationInstances(spec.getClass()));

						if(annotationFilter.passesColumnFilter(columnHandle)) {
							sortedColumnHandles.add(columnHandle);
						}

					}
				}
			}

			// Instantiate columns
			for(ColumnHandle columnHandle : sortedColumnHandles) {
				createAnnotationColumn(columnHandle);
			}

			tableViewer.getTable().setRedraw(true);
		}

		// Update the view
		tableViewer.refresh();
	}

	/** {@inheritDoc} */
	@Override
	protected void update(IModelElement modelElement) {
		// Updates the contents of the table
		super.update(modelElement);

		// ensure that selected model element is visible in annotation view
		AnnotationEntry entry = IAnnotationValueService.INSTANCE.getAnnotationEntry(modelElement);
		assert (entry.getModelElement() == modelElement);

		tableViewer.reveal(entry);
	}

	/** Initializes the {@link AnnotationFilterWidget}. */
	private void setupAnnotationFilterWidget() {
		annotationFilterWidget = new AnnotationFilterWidget(rootComposite, SWT.NULL, this);
		annotationFilterWidget.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
	}

	/** {@inheritDoc} */
	@Override
	public void createPartControl(Composite parent) {
		super.createPartControl(parent);
		rootComposite = new Composite(parent, SWT.NULL);
		rootComposite.setLayout(new GridLayout(1, false));

		tableViewer = new TableViewer(rootComposite, SWT.BORDER | SWT.FULL_SELECTION);
		Table table = tableViewer.getTable();
		table.setHeaderVisible(true);
		table.setLinesVisible(true);
		table.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true, 1, 1));

		ColumnViewerToolTipSupport.enableFor(tableViewer, ToolTip.NO_RECREATE);

		// Install layout for Table in order to ensure that it claims all available vertical space
		// in case the
		// filter option GUI is removed (see setFixedAnnotationFilter()).
		table.setLayout(new FillLayout(SWT.VERTICAL));

		fixedColumnCount = createFixedModelElementColumns();
		columnWidthCache = new HashMap<String, Integer>();

		tableViewer.setComparator(new HierarchicalNameComparator());

		tableViewer.setComparer(new IElementComparer() {

			@Override
			public int hashCode(Object element) {
				if(element instanceof AnnotationEntry) {
					return ((AnnotationEntry)element).getModelElement().hashCode();
				}
				return element.hashCode();
			}

			@Override
			public boolean equals(Object a, Object b) {
				if(a instanceof AnnotationEntry && b instanceof AnnotationEntry) {
					return ((AnnotationEntry)a).getModelElement() == ((AnnotationEntry)b)
							.getModelElement();
				}
				return a.equals(b);
			}
		});

		// Rows (= components) are filtered using a specialized ViewerFilter.
		// Since the data model is an array of AnnotationEntries, columns (= annotations)
		// cannot be filtered using such a filter, and require the reconstruction of the
		// table, with an implementation of the filter in update(Set<AnnotationEntry>).
		annotationFilter = new AnnotationFilter();
		tableViewer.addFilter(annotationFilter);

		// Add filter widget
		setupAnnotationFilterWidget();
	}

	/** Creates a fixed (leading) column to be shown for every model element */
	private void createFixedColumn(String label, ColumnLabelProvider labelProvider,
			EditingSupport editingSupport) {
		TableViewerColumn column = new TableViewerColumn(tableViewer, SWT.NONE);
		column.getColumn().setText(label);
		column.getColumn().setWidth(COLUMN_DEFAULT_WIDTH);

		tableViewer.setContentProvider(new ArrayContentProvider());
		column.setLabelProvider(labelProvider);
		column.setEditingSupport(editingSupport);
	}

	/**
	 * Creates the (leading) columns which displays the model elements' names and comments.
	 * 
	 * @returns Number of leading columns
	 * */
	protected int createFixedModelElementColumns() {
		createFixedColumn("Model Element", new ElementNameLabelProvider(this),
				new ElementNameEditingSupport(tableViewer));

		createFixedColumn("Comment", new ElementCommentLabelProvider(this),
				new ElementCommentEditingSupport(tableViewer));

		return 2;
	}

	/** Creates a column in tableViewer that displays the annotation given in spec */
	protected void createAnnotationColumn(ColumnHandle columnHandle) {

		TableColumn tableColumn = null;
		if(columnHandle.isCreateColumn()) {
			// Create column that contains a button for adding new instances of the
			// current annotation type.

			CreateAnnotationInstanceColumn column =
					new CreateAnnotationInstanceColumn(tableViewer, SWT.NONE,
							columnHandle.getAnnotatedSpecification(), columnHandle.getEntry());
			tableColumn = column.getColumn();

		} else {
			Class<? extends IAnnotatedSpecification> annotationClass =
					columnHandle.getAnnotatedSpecification().getClass();

			// Add new column
			try {
				TableViewerColumn column =
						createAnnotationTableViewerColumn(columnHandle, annotationClass);

				tableColumn = column.getColumn();
				String specName = columnHandle.getColumnName();
				tableColumn.setText(specName);
			} catch(Exception e) {
				System.out.println("Failed to create editing support for annotation class " +
						annotationClass.getCanonicalName() + ": " + e.getMessage());
			}

		}

		if(tableColumn != null) {
			Integer width = columnWidthCache.get(tableColumn.getText());
			if(width != null) {
				// If possible, use cached with (possibly adjusted by the user)
				tableColumn.setWidth(width);
			} else {
				// Otherwise, calculate width for new columns
				tableColumn.pack();
				width = tableColumn.getWidth();
				if(width < COLUMN_DEFAULT_WIDTH) {
					tableColumn.setWidth(COLUMN_DEFAULT_WIDTH);
				} else if(width > 1.2f * COLUMN_DEFAULT_WIDTH) {
					tableColumn.setWidth(Math.round(1.2f * COLUMN_DEFAULT_WIDTH));
				}
			}
		}

	}

	/**
	 * Creates a {@link TableViewerColumn}, whose {@link EditingSupport} and {@link LabelProvider}
	 * is subject to the provided return value of the given {@link IAnnotatedSpecification}.
	 * If something else than pure text shall be displayed for the annotation, these extended
	 * {@link LabelProvider} shall be instantiated here.
	 */
	private TableViewerColumn createAnnotationTableViewerColumn(ColumnHandle columnHandle,
			Class<? extends IAnnotatedSpecification> annotationClass) throws Exception {
		TableViewerColumn column = new TableViewerColumn(tableViewer, SWT.NONE);

		// Have the matching EditingSupport created for the current annotated specification
		// (work is delegated to respective IAnnotationValueProvider implementation)
		EditingSupport editingSupport =
				columnHandle.getEntry().createSpecificationEditElement(tableViewer,
						annotationClass, columnHandle.getInstanceKey());
		ColumnLabelProvider columnLabelProvider = null;

		// TODO: Avoid this case distinction here and delegate the construction of label providers
		// to the value provider (see #2226).
		if(columnHandle.getEntry().getSpecificationValue(annotationClass) instanceof Boolean) {
			columnLabelProvider =
					new EmulatedNativeCheckBoxLabelProvider(annotationClass, this,
							columnHandle.getInstanceKey(), column.getViewer());
		} else {
			columnLabelProvider =
					new AnnotationLabelProvider(annotationClass, this,
							columnHandle.getInstanceKey(), editingSupport);
		}
		column.setEditingSupport(editingSupport);
		column.setLabelProvider(columnLabelProvider);

		return column;
	}

	/** {@inheritDoc} */
	@Override
	public void setFocus() {
		// nothing to do
	}

	/** {@inheritDoc} */
	@Override
	public void dispose() {
		tableViewer.getTable().dispose();
		super.dispose();
	}

	/** Returns lastAnnotationEntries. */
	public Collection<AnnotationEntry> getLastAnnotationEntries() {
		return lastAnnotationEntries;
	}
}
